
import ij.ImagePlus;
import ij.gui.Plot;
import ij.gui.Roi;
import ij.measure.ResultsTable;
import ij.plugin.DICOM;
import ij.plugin.PlugIn;
import ij.plugin.filter.GaussianBlur;
import ij.process.AutoThresholder;
import ij.process.Blitter;
import ij.process.ImageProcessor;
import ij.process.ImageStatistics;
import ij.util.DicomTools;
import ij.*;
import ij.gui.*;
import ij.io.DirectoryChooser;
import java.awt.*;
import java.io.File;
import java.util.ArrayList;

/**
 * Auxiliary Application for Mammography system calibration and testing
 * Should be used as ImageJ plug-in
 * If you choose full mode ,  make more memory available to ImageJ, see
 * http://rsbweb.nih.gov/ij/docs/install/windows.html#memory
 * Expected image format: DICOM
 * This application supports three image quality tests:
 * 1. Contrast to Noise Ratio (CNR)
 * 2. Detector ghosting (SDNR) 
 * 3. Modulation transfer function(MTF)
 * @author alex
 *
 */
public class MCalibrPlugIn_ implements PlugIn {
	private boolean visible = false;
	
	public void run(String arg) {	
		GenericDialog gd = new GenericDialog("Choose task");
		String[] labels = {"CNR", "Detector ghosting", "MTF"};
		gd.addChoice("Choose task", labels, "CNR");
		
		gd.addCheckbox("Full mode : view intermediate images for CNR/Ghost tests", false);
		gd.showDialog();

		DirectoryChooser chooser = new DirectoryChooser("Choose a Directory");
		String path = chooser.getDirectory();
		int choice = gd.getNextChoiceIndex();
		if(gd.getNextBoolean()){
			IJ.showMessage("Full mode is memory consuming - make more memory available to ImageJ\n" +
					"see http://rsbweb.nih.gov/ij/docs/install/windows.html#memory");
			visible = true;
		}
		if(choice == 0)
			chooseCNR(path);
		else if (choice == 1)
			chooseGhost(path);
		else
			chooseMTF(path);
	}
	
	/*
	 * Contrast to Noise Ratio (CNR) test
	 * @param directory with input images
	 */
	private void chooseCNR(String path){
		File actual = new File(path);
		if(actual.listFiles().length != 6){
			IJ.error("CNR test: Directory " + path + " does not contain 6 DICOM files");
			throw new RuntimeException("Directory " + path + " does not contain 6 DICOM files");
		}
		
		DICOM dcmCreat = new DICOM();
		String currPath;
		ArrayList<Double> CNRs = new ArrayList<Double>();
		ArrayList<Double> times = new ArrayList<Double>();
		int[] limits = {115, 110, 105, 100, 95, 90};
		int[] thicknessPMMA = {20, 30, 40, 50, 60, 70};
		double tmpCNR = 0;
		double tmpTime = 0;
        for( File f : actual.listFiles()){
        	currPath = path+f.getName();
        	dcmCreat.open(currPath);
            tmpTime = Double.parseDouble(DicomTools.getTag(dcmCreat, "0008,0033"));
            tmpCNR = calculateCNR(currPath);
            int i;
            for(i = 0; i < times.size(); i++){
            	if(times.get(i) > tmpTime){
            		times.add(i, tmpTime);
            		CNRs.add(i, tmpCNR);
            		break;
            	}
            }
            if(i == times.size()){
            	times.add(tmpTime);
        		CNRs.add(tmpCNR);
            }
        }
        
		ResultsTable rt = new ResultsTable();
		for (int i = 0; i < 6; i++) {
			rt.incrementCounter();
			rt.addValue("CNR absolute", CNRs.get(i));
			rt.addValue("Limiting values, %", limits[i]);
			rt.addValue("CNR relative to 50mm, %", (int)(100*CNRs.get(i)/CNRs.get(3)));
			rt.addValue("PMMA thickness, mm", thicknessPMMA[i]);
			
		}
		rt.showRowNumbers(false);
		rt.show("CNR test results");    
	}
	
	/*
	 * Detector ghosting (SDNR) 
	 * @param directory with input images
	 */
	private void chooseGhost(String path){
		double limit = 2.0;
		File actual = new File(path);
		if(actual.listFiles().length != 2){
			IJ.error("ghost test: Directory " + path + " does not contain 2 DICOM files");
			throw new RuntimeException("Directory " + path + " does not contain 2 DICOM files");
		}
		String path1 = path+actual.listFiles()[0].getName();
		String path2 = path+actual.listFiles()[1].getName();
		DICOM dcmCreat = new DICOM();
		dcmCreat.open(path1);
		double time1 = Double.parseDouble(DicomTools.getTag(dcmCreat, "0008,0033"));
		dcmCreat.open(path2);
		double time2 = Double.parseDouble(DicomTools.getTag(dcmCreat, "0008,0033"));
		double SDNR;
		if(time1 < time2)
			SDNR = calculateGhost(path1, path2);
		else
			SDNR = calculateGhost(path2, path1);
        
		ResultsTable rt = new ResultsTable();
		rt.incrementCounter();
		rt.addValue("SDNR", SDNR);
		rt.addValue("Limiting values", limit);
		rt.showRowNumbers(false);
		rt.show("detector ghosting test results");   
	}
	
	/*
	 * Modulation transfer function(MTF)
	 * @param directory with input image 
	 */
	private void chooseMTF(String path){
		File actual = new File(path);
		if(actual.listFiles().length != 1){
			IJ.error("MTF test: Directory " + path + " does not contain 1 DICOM files");
			throw new RuntimeException("Directory " + path + " does not contain 1 DICOM files");
		}
		String fullpath = path+actual.listFiles()[0].getName();
		calculateMTF(fullpath);
	}
	
	/*
	 * MTF computation main procedure
	 * @param input image absolute path
	 */
	private void calculateMTF(String path){
		DICOM dcm = new DICOM();
        dcm.open(path);
        int testWindowSizePix = getTestWindowSizePix(dcm);
        
        ImageProcessor ip = dcm.getProcessor();
        int[] edges = getEdgeROI(path);
        int verticalEdgeX = edges[0];
        int verticalEdgeY = edges[1];
        int horizontalEdgeX = edges[2];
	    int horizontalEdgeY = edges[3];
	    
	    //vertical edge test
	    ImageProcessor newIMP;
	    ImagePlus newIM;
	    Roi newRoi;
	    if(verticalEdgeX > horizontalEdgeX){
		    newIMP = (ip.rotateRight()).rotateRight();		//rotate image 180 deg right
		    newIM =  new ImagePlus("MTF image, vertical edge test", newIMP);
		    newRoi = new Roi((newIMP.getWidth() - verticalEdgeX) - testWindowSizePix/2,
		    		(newIMP.getHeight() - verticalEdgeY) - testWindowSizePix/2,	testWindowSizePix, testWindowSizePix);
	    }
	    else{
	    	newIM = dcm;
	    	newRoi = new Roi(verticalEdgeX - testWindowSizePix/2, verticalEdgeY - testWindowSizePix/2,
		    		testWindowSizePix, testWindowSizePix);
	    }
	    newIM.setRoi(newRoi);
	    newIM.show();
	    newIM.draw();

	    //MTF module call - vertical and horizontal edge
	    MTFModule MTFcalculatorVertical = new MTFModule(newIM, getTestWindowSizeMM(dcm));
		MTFcalculatorVertical.run("vertical");
		
		//horizontal edge test
	    newIMP = ip.rotateRight();
	    newIM =  new ImagePlus("MTF image, horizontal edge test", newIMP);
	    newRoi = new Roi(newIMP.getWidth() - (horizontalEdgeY - testWindowSizePix/2) - testWindowSizePix
		, horizontalEdgeX - testWindowSizePix/2, testWindowSizePix, testWindowSizePix);
	    newIM.setRoi(newRoi);
	    newIM.show();
	    newIM.draw();
	    
	    MTFModule MTFcalculatorHorizontal = new MTFModule(newIM, getTestWindowSizeMM(dcm));
		MTFcalculatorHorizontal.run("horizontal");
	}
	
	/*
	 * computes CNR
	 * given 6 DICOM files containing standard CNR test images 
	 * @param directory with input images 
	 * @return CNR
	 */
	private double calculateCNR(String filePath){
		DICOM dcm = new DICOM();
        dcm.open(filePath);
        ImageProcessor myProcessor = dcm.getProcessor();
        int testWindowSizePix = getTestWindowSizePix(dcm);
        
        //[verticalEdgeX, verticalEdgeY, horizontalEdgeX, horizontalEdgeY]
        int[] edges = getEdgeROI(filePath);   
        int verticalEdgeY = edges[1];
        int horizontalEdgeX = edges[2];
	    int horizontalEdgeY = edges[3];

	    int alumRoiCenterX = horizontalEdgeX;
	    int alumRoiCenterY = verticalEdgeY;
	    
	    int PMMARoiCenterX = horizontalEdgeX;
	    int PMMARoiCenterY = 2*horizontalEdgeY-verticalEdgeY;
        
	    //calculate mean and standard deviation for each ROI
	    Roi newRoi = new Roi(PMMARoiCenterX - testWindowSizePix/2, PMMARoiCenterY - testWindowSizePix/2,
	    		testWindowSizePix, testWindowSizePix);
        myProcessor.setRoi(newRoi);
        
        ImageProcessor myProcessorPMMA = myProcessor.crop();
        ImagePlus dcmPMMA = new ImagePlus("PMMA ROI image", myProcessorPMMA);
        ImageStatistics PMMAstats = dcmPMMA.getStatistics();
        double PMMAmean = PMMAstats.mean;
        double PMMAstdDev = PMMAstats.stdDev;
        //show intermediate images
        if(visible){
        	dcmPMMA.show();
            dcmPMMA.draw();
            
            myProcessor.setValue(500);
            myProcessor.fill(newRoi);
        }

	    //calculate mean and standard deviation for each ROI
	    newRoi = new Roi(alumRoiCenterX - testWindowSizePix/2, alumRoiCenterY - testWindowSizePix/2,
	    		testWindowSizePix, testWindowSizePix);
        myProcessor.setRoi(newRoi);
        
        ImageProcessor myProcessoralum = myProcessor.crop();
        ImagePlus dcmalum = new ImagePlus("alum ROI image", myProcessoralum);
        ImageStatistics alumstats = dcmalum.getStatistics();
        double alummean = alumstats.mean;
        double alumstdDev = alumstats.stdDev;
        //show intermediate images
        if(visible){
	        dcmalum.show();
	        dcmalum.draw();
	        myProcessor.setValue(500);
	        myProcessor.fill(newRoi);
	        dcm.show();
	        dcm.draw();
        }
        
        //calculate CNR
        double CNR = (alummean - PMMAmean)/Math.sqrt((Math.pow(alumstdDev, 2) + Math.pow(PMMAstdDev, 2))/2);    
        return CNR;
	}
	
	/*
	 * calculate ROI size in mm
 	 * the result depends on distance between pixels in DICOM image
	 */
	private double getTestWindowSizeMM(DICOM dcm){
		String pixeDist = DicomTools.getTag(dcm, "0028,0030");	//extract pixel width field from DICOM header
		return Double.parseDouble(pixeDist.substring(0, pixeDist.indexOf('\\')));  
	}
	
	/*
	 * calculate ROI size in pixels
 	 * the result depends on distance between pixels in DICOM image
	 */
	private int getTestWindowSizePix(DICOM dcm){
		String pixeDist = DicomTools.getTag(dcm, "0028,0030");	//extract pixel width field from DICOM header
	    double pixelWidth = Double.parseDouble(pixeDist.substring(0, pixeDist.indexOf('\\')));  
	    double testWindowSizeMM = 20; // 20mm = 2cm
	    return (int)Math.ceil(testWindowSizeMM / pixelWidth);
	}
	
	/*
	 * calculates coordinates of two points, one from vertical and one from horizontal edge
	 * @return [verticalEdgeX, verticalEdgeY, horizontalEdgeX, horizontalEdgeY]
	 */
	private int[] getEdgeROI(String filePath){
		DICOM dcm = new DICOM();
        dcm.open(filePath);
		ImageProcessor myProcessor = dcm.getProcessor();	
		
	    DICOM dcmMeasureCopy = new DICOM();
	    dcmMeasureCopy.open(filePath);
        int[] edges = getBounds(dcmMeasureCopy, AutoThresholder.Method.Default);
        
        //convert array indices to pixel coordinates
        int westBound = edges[0]; 
        int eastBound = edges[1]; 
        int northBound = edges[2]; 
        int southBound = edges[3]; 
                
        int infoImWidth = eastBound - westBound;
        int infoImHeight = southBound - northBound;
        
        int informativeImageCenter_X = westBound + infoImWidth/2;
        int informativeImageCenter_Y = northBound + infoImHeight/2;
        
        int borderCut = 250;
        westBound = informativeImageCenter_X - infoImWidth/2 + borderCut; 
        eastBound = informativeImageCenter_X + infoImWidth/2 - borderCut;
        northBound = informativeImageCenter_Y - infoImHeight/2 + borderCut;
        southBound = informativeImageCenter_Y + infoImHeight/2 - borderCut;
        
        //coordinate conversion factor: from cropped image to real image  (x, y) => (x+deltaX, y+deltaY)
        int deltaX = westBound;
        int deltaY = northBound;

        Roi newRoi = new Roi(westBound, northBound,
        		eastBound - westBound, southBound - northBound);
        myProcessor.setRoi(newRoi);
        
        ImageProcessor myProcessorCropped = myProcessor.crop();
        ImagePlus dcmCropped = new ImagePlus("cropped image", myProcessorCropped);
        
        edges = getBounds(dcmCropped, AutoThresholder.Method.Default);
        //convert array indices to pixel coordinates
        westBound = edges[0]; 
        eastBound = edges[1]; 
        northBound = edges[2]; 
        southBound = edges[3]; 
        
        int verticalEdgeX;
        int verticalEdgeY = northBound + (southBound - northBound)/2;
        int horizontalEdgeX = westBound + (eastBound - westBound)/2;
	    int horizontalEdgeY;  
        if(westBound == 0){
        	verticalEdgeX = eastBound;
        }
        else{
        	verticalEdgeX = westBound;
        }
        if(northBound == 0){
        	horizontalEdgeY = southBound;
        }
        else{
        	horizontalEdgeY = westBound;
        }

	    myProcessor.resetRoi();		//clear ROI selection
	    int[] result  = {verticalEdgeX+deltaX, verticalEdgeY+deltaY, horizontalEdgeX+deltaX, horizontalEdgeY+deltaY};
	    return result; 
	}
	
	/*
	 * calculates four bounds of the image (removes background)
	 * @return [westBound, eastBound, northBound, southBound]
	 */
	private int[] getBounds(ImagePlus dcm, AutoThresholder.Method type){
		ImageProcessor myProcessor = dcm.getProcessor();
		int w = dcm.getWidth();
	    int h = dcm.getHeight();
	    int verticalMiddle = h/2;
	    int horizontalMiddle = w/2;
		
		//calculate minimum threshold for background filtering
	    AutoThresholder thresholder = new AutoThresholder();
	    int smallTr = thresholder.getThreshold(type, myProcessor.getHistogram());
	
	    //create filtered copies of the images (image = 255, background = 0)
	    GaussianBlur filterr = new GaussianBlur();
	    double radiuss = 6.0;
	    
	    filterr.blurGaussian(myProcessor, 0.4*radiuss, 0.4*radiuss, 0.01);	//filter noise with gaussian
	    myProcessor.threshold(smallTr);
	    
	    short[] pixels = (short[]) myProcessor.getPixels();
		
	    //find image bounds
	    int westBound = w*verticalMiddle;
	    int until = w*(verticalMiddle + 1);
	    while(westBound < until && pixels[westBound] == 0) {
	    	westBound++;
	    }
	    if(westBound > (w*verticalMiddle + horizontalMiddle))
	    	westBound = w*verticalMiddle;
	    
	    int eastBound = w*(verticalMiddle + 1) - 1;
	    until = w*verticalMiddle;
	    while(eastBound > until && pixels[eastBound] == 0) {
	    	eastBound--;
	    }
	    if(eastBound < (w*verticalMiddle + horizontalMiddle))
	    	eastBound = w*(verticalMiddle + 1) - 1;
	    
	    int northBound = horizontalMiddle;
	    until = w*(h - 1) + horizontalMiddle;
	    while(northBound < until && pixels[northBound] == 0) {
	    	northBound += w;
	    }
	    
	    int southBound = w*(h - 1) + horizontalMiddle;
	    until = horizontalMiddle;
	    while(southBound > until && pixels[southBound] == 0) {
	    	southBound -= w;
	    }
	    
	    //convert array indices to pixel coordinates
	    westBound %= w; 
	    eastBound %= w;
	    northBound /= w;
	    southBound /= w;
	    
	    int[] result  = {westBound, eastBound, northBound, southBound};
	    return result;
	}
	
	/*
	 * given ghost creation and ghost measurement files compute detector ghosting
	 * @param ghost creation file path
	 * @param ghost measurement file path
	 * @return SDNR
	 */
	private double calculateGhost(String ghostCreatPath, String ghostMeasurePath){
		DICOM dcmCreat = new DICOM();
		dcmCreat.open(ghostCreatPath);
		
        DICOM dcmMeasure = new DICOM();
        dcmMeasure.open(ghostMeasurePath);
        ImageProcessor Measureip = dcmMeasure.getProcessor();
        
        int testWindowSizePix = getTestWindowSizePix(dcmCreat);

        //calculate minimum threshold for background filtering
        AutoThresholder thresholder = new AutoThresholder();
        int smallTr = thresholder.getThreshold(AutoThresholder.Method.Default, Measureip.getHistogram());

        //create filtered copies of the images (image = 255, background = 0)
        GaussianBlur filter = new GaussianBlur();
        double radius = 6.0;
        
        DICOM dcmMeasureCopy = new DICOM();
        dcmMeasureCopy.open(ghostMeasurePath);
        ImageProcessor MeasureipCopy = dcmMeasureCopy.getProcessor();
        filter.blurGaussian(MeasureipCopy, 0.4*radius, 0.4*radius, 0.01);	//filter noise with gaussian
        MeasureipCopy.threshold(smallTr);
        
        DICOM dcmCreatCopy = new DICOM();
        dcmCreatCopy.open(ghostCreatPath);
        ImageProcessor CreatipCopy = dcmCreatCopy.getProcessor();
        filter.blurGaussian(CreatipCopy, 0.4*radius, 0.4*radius, 0.01);		//filter noise with gaussian
        CreatipCopy.threshold(smallTr);

        //merge filtered image copies
        CreatipCopy.copyBits(MeasureipCopy, 0, 0, Blitter.ADD);
        
        //find attenuated ROI (intersection of two images)
        int w = dcmCreatCopy.getWidth();
        int h = dcmCreatCopy.getHeight();
        int imageCenterX = w/2;
	    
	    short[] pixels= (short[]) CreatipCopy.getPixels();
	    int northBound = imageCenterX;
	    int southBound = w*(h - 1) + imageCenterX;
	    
	    int i;
	    for(i = imageCenterX; i <= w*(h - 1) + imageCenterX; i += w){
	    	if(pixels[i] >= 255*2){		//filtered image, image = 255, background = 0
	    		northBound = i/w;
	    		break;
	    	}
	    }
	    for(i = w*(h - 1) + imageCenterX; i > imageCenterX; i -= w){
	    	if(pixels[i] >= 255*2){		//filtered image, image = 255, background = 0
	    		southBound = i/w;
	    		break;
	    	}	
	    }
	    
	    int westBound = w*((southBound - testWindowSizePix)-1) + imageCenterX;
	    int eastBound = w*((southBound - testWindowSizePix)-1) + imageCenterX;

	    for(i = w*((southBound - testWindowSizePix)-1) + imageCenterX; i >= w*((southBound - testWindowSizePix) - 1); i--){
	    	if(pixels[i] < 255*2){		//filtered image, image = 255, background = 0
	    		westBound = i%w + 1;
	    		break;
	    	}	
	    }
	    for(i = w*((southBound - testWindowSizePix)-1) + imageCenterX; i < w*(southBound - testWindowSizePix) ; i++){
	    	if(pixels[i] < 255*2){		//filtered image, image = 255, background = 0
	    		eastBound = i%w - 1;
	    		break;
	    	}	
	    }
	    
	    int attinuatedRoiCenterX = westBound + (eastBound - westBound)/2;
	    int attinuatedRoiCenterY = northBound + (southBound - northBound)/2;
 
	    //find unattenuated ROI
	    int unattinuatedRoiCenterX = attinuatedRoiCenterX;
	    int unattinuatedRoiCenterY = attinuatedRoiCenterY;
	    pixels= (short[]) MeasureipCopy.getPixels();
	    int upCheckPoint  = w*(northBound - 1 - testWindowSizePix) + attinuatedRoiCenterX;
	    int downCheckPoint  = w*(southBound - 1 + testWindowSizePix) + attinuatedRoiCenterX;
	    
	    if(pixels[upCheckPoint] > 0){
		    for(i = upCheckPoint; i >= attinuatedRoiCenterX; i -= w){
		    	if(pixels[i] < 255){		//filtered image, image = 255, background = 0
		    		unattinuatedRoiCenterY = (i/w + 1) + (northBound - (i/w + 1))/2; 
		    		break;
		    	}
		    }
	    }
	    else{
	    	for(i = downCheckPoint; i < (w*(h-1) + attinuatedRoiCenterX); i += w){
		    	if(pixels[i] < 255){		//filtered image, image = 255, background = 0
		    		unattinuatedRoiCenterY = southBound + ((i/w - 1) - southBound)/2; 
		    		break;
		    	}
		    }
	    }
	    
	    //calculate mean and standard deviation for attenuated ROI
	    Roi newRoi = new Roi(attinuatedRoiCenterX - testWindowSizePix/2, attinuatedRoiCenterY - testWindowSizePix/2,
	    		testWindowSizePix, testWindowSizePix);
	    Measureip.setRoi(newRoi);
        
        ImageProcessor myProcessorattinuated = Measureip.crop();
        ImagePlus dcmattinuated = new ImagePlus("attinuated ROI image", myProcessorattinuated);
        ImageStatistics attinuatedstats = dcmattinuated.getStatistics();
        double attinuatedmean = attinuatedstats.mean;
        double attinuatedStdDev = attinuatedstats.stdDev;
        //show intermediate images
        if(visible){
	        dcmattinuated.show();
	        dcmattinuated.draw();
	        
	        Measureip.setValue(500);
	        Measureip.fill(newRoi);
        }

        //calculate mean and standard deviation for unattenuated ROI
        newRoi = new Roi(unattinuatedRoiCenterX - testWindowSizePix/2, unattinuatedRoiCenterY - testWindowSizePix/2,
	    		testWindowSizePix, testWindowSizePix);
	    Measureip.setRoi(newRoi);
        
        ImageProcessor myProcessorunattinuated = Measureip.crop();
        ImagePlus dcmunattinuated= new ImagePlus("unattinuated ROI image", myProcessorunattinuated);
        ImageStatistics unattinuatedstats = dcmunattinuated.getStatistics();
        double unattinuatedmean = unattinuatedstats.mean;
        //show intermediate images
        if(visible){
        	dcmunattinuated.show();
        	dcmunattinuated.draw();   
        	Measureip.setValue(500);
        	Measureip.fill(newRoi);
        	dcmMeasure.show();
        	dcmMeasure.draw();
        }
        
        double CDNR = Math.abs((unattinuatedmean - attinuatedmean)/attinuatedStdDev);
        return CDNR;        
	}
	
	/**
	 * MTF calculation module
	 * calculates values and plot graphs
	 * adaptation of general MTF implementation from 
	 * http://rsbweb.nih.gov/ij/index.html
	 */
	private class MTFModule{
		final private int nPhotodetectors = 1;
		final private int frameSize = 32;		//test frame size - 32 pixels
		final private double ny;		//scale unit
		private ImagePlus imp;
		private Roi roi;
		private int imageWidth;
		private int imageHeight;
		private double[][] PosMax;
		private double[][] ESFArray;
		private double[][] LSFArray;

		/**
		 * Constructor
		 * @param imp
		 * @param pixelWidth
		 */
		public MTFModule(ImagePlus imp, double pixelWidth){
			this.imp = imp;
			this.ny = (nPhotodetectors/pixelWidth);
			this.roi = imp.getRoi();	
			if (roi==null) {
				imp.setRoi(0, 0, imp.getWidth(), imp.getHeight());
				roi = imp.getRoi();
				IJ.showMessageWithCancel("Warning","All image selected");
        	}
			this.imageWidth=roi.getBounds().width;
			this.imageHeight=roi.getBounds().height;
		}
		
		/**
		 * Module start function
		 * @param arg
		 */
		public void run(String arg){
	    	double[]ESFVector;
			double[]LSFVector;
			double[]LSFDVector;
			double[]MTFVector;
			generateESFArray("ESF Plot",imp,roi);
			generateLSFArray("LSF Plot",ESFArray);
			calculateMax();
			
			double[][] ESFArrayF=alignArray(ESFArray);
			double[][] LSFArrayF=alignArray(LSFArray);
			ESFVector=averageVector(ESFArrayF);
			LSFVector=averageVector(LSFArrayF);
			int aura = (LSFVector.length * 2);
			LSFDVector = new double [aura];
			int j = 0;	
			int i;
			for(i=0;i<(LSFDVector.length-3); i++){
				
				if(i % 2 == 0) { 
					LSFDVector[i]= LSFVector[j];
					j=j+1;
				}else {
					LSFDVector[i]= ((0.375*LSFVector[(j-1)]) + (0.75*LSFVector[(j)]) - (0.125*LSFVector[(j+1)]));
				}
			} 		
			LSFDVector[i] = ((LSFVector[j-1] + LSFVector[j])*0.5);
			LSFDVector[i+1] = LSFVector[j];
			LSFDVector[i+2] = LSFVector[j];
	
			int indexMax = 0;
			double valorMax = LSFDVector[0];
			for(i=0;i<LSFDVector.length;i++){
				  if(valorMax < LSFDVector[i]){
			      indexMax = i;
			      valorMax = LSFDVector[i];
			   }
			}
			i=indexMax;
			LSFDVector[i-1]=((LSFDVector[i-2] + LSFDVector[i])*0.5);
			
			MTFVector=fftConversion(LSFDVector, "MTF");
			String title=imp.getTitle();
			generatePlot(MTFVector,"MTF", title);
			generatePlot(LSFVector,"LSF", title);
			generatePlot(ESFVector,"ESF", title);
			//generatePlot(SPPVector,"SPP");
			cleanImage();
	    }			
		
		private void cleanImage(){		
				imp.killRoi();
		}
			
		
		/**
		 * Step function generation procedure
		 * @param title
		 * @param imp
		 * @param roi
		 */
		private void generateESFArray(String title, ImagePlus imp, Roi roi) {	
			double[] ESFLinea;
			ProfilePlot plotESF;
			Rectangle r=roi.getBounds();					
			
			if (frameSize>=imageWidth){
				IJ.showMessage("Error","sample size is bigger than selection width\nProcess canceled");
				return;
			}
			
			int selectX=r.x;
			int selectY=r.y;
			int selectXFin=selectX+imageWidth;
			ESFLinea=new double[imageWidth];
			ESFArray=new double[imageHeight][imageWidth];
	        for(int k=0;k<imageHeight;k++){
				//select line 
				IJ.makeLine(selectX,k+selectY,selectXFin-1,k+selectY);
				//save values 
				plotESF = new ProfilePlot(imp);
				ESFLinea=plotESF.getProfile();
				//Load Array ESF
				for(int i=0;i<imageWidth;i++){
					ESFArray[k][i]=ESFLinea[i];
				}
			}
	    }
		
		/**
		 * Line Spread Function -LSF generation procedure
		 * @param title
		 * @param ESFArray
		 */
		private void generateLSFArray(String title, double[][]ESFArray){
			
			LSFArray=new double[imageHeight][imageWidth];		
			
			for(int k=0;k<imageHeight;k++){
				for(int i=0;i<imageWidth-1;i++){
					LSFArray[k][i]=ESFArray[k][i+1]-ESFArray[k][i];
				}
			}
		}
		
		/**
		 * slanted edge alignment procedure
		 * @param Array
		 * @return
		 */
		private double[][] alignArray(double[][] Array){
			double[][] ArrayF = new double[imageHeight][frameSize];
			int ini;
			int fin;
			
			//Create new array aligned
			for(int k=0;k<imageHeight;k++){
				
				//Initial and end positions of k-line
				ini=(int)PosMax[k][2];
				fin=(int)PosMax[k][3];
				for(int i=ini;i<fin;i++){	
					ArrayF[k][i-ini]=Array[k][i];
				}
			}
			return ArrayF;
		}
		

		//Calculate maximum value and find 32 positions to align
		private void calculateMax(){
			
			PosMax=new double[imageHeight][4] ;
			int posMax;
			int halfSize=frameSize/2;
			for(int k=0;k<imageHeight-1;k++){
				posMax=0;
				for(int i=0;i<imageWidth-1;i++){
					if (LSFArray[k][posMax]<LSFArray[k][i] || LSFArray[k][posMax]==LSFArray[k][i]){
						posMax=i;
					}
				}

				if((posMax + halfSize) > roi.getBounds().getWidth() ||
						(posMax - halfSize) < 1)
					posMax = (int)PosMax[k-1][0];
				//The maximum value position on the line
				PosMax[k][0]=posMax;
				//The maximum value on the line
				PosMax[k][1]=LSFArray[k][posMax];
				//Starting and ending position to align maximum values
				PosMax[k][2]=PosMax[k][0]-halfSize;
				PosMax[k][3]=PosMax[k][0]+halfSize;
			}
		}
		
		
		private double[] averageVector(double[][] Array){
			
			double result;
			double[]Vector = new double[frameSize];
			
			//Average of all linear ESF/LSF 	
			for(int i=0;i<frameSize;i++){
				result=0;
				//Average of all rows i-position  
				for( int k=0;k<imageHeight;k++){	
					result=result+Array[k][i];	
				}	
				result=result/imageHeight;
				Vector[i]=result;
			}
			return Vector;
		}
								
		private void generatePlot(double[] Vector, String plot, String title){
			double[]xValues;
			String ejeX="pixel";
			String ejeY="";
			String allTitle="";
			int yMax = 2000;		//experimental constant, produce readable plot (instead of 65535)
			
			
			//If MTF plot, calculate the scale of cycles per pixel for x-axis values
			xValues=calculateXValues(Vector,plot);
			
			//plot titles 		
			if (plot=="ESF"){
				ejeY="Grey Value";
			}

			if (plot=="LSF"){ 
				ejeY="Grey Value / pixel";
			}
			
			if (plot=="MTF"){
				ejeY="Modulation Factor";
				ejeX ="lp/mm";
			}
			
			if (plot=="SPP"){
				ejeY="SPP";
			}

			allTitle=plot + "_" + title;	
			Plot plotResult = new Plot(allTitle, ejeX, ejeY, xValues, Vector);
			
			//plot limits
			if (plot=="ESF"){
				plotResult.setLimits(1,Vector.length,0,yMax);
			}
			
			if (plot=="LSF"){
				plotResult.setLimits(1,Vector.length,0,yMax);
			}
			
			if (plot=="MTF"){
				plotResult.setLimits(0,ny,0,1);
			}
					
			if (plot=="SPP"){
				plotResult.setLimits(1,Vector.length,0,1);
			}
			 
			plotResult.draw();
			plotResult.show();
			
			//display MTF result table, according to standard
			if(plot=="MTF"){
				int i;
				double x50, x20;
				for(i = 0; (i < xValues.length) && (Vector[i] > 0.5); i++){}
				x50 = xValues[i];
				for(; (i < xValues.length) && (Vector[i] > 0.2); i++){}
				x20 = xValues[i];

				ResultsTable rt = new ResultsTable();
				rt.incrementCounter();
				rt.addValue("fall of MTF below 50%,\n frequency (lp/mm)", x50);
				rt.addValue("fall of MTF below 20%,\n frequency (lp/mm)", x20);
				rt.showRowNumbers(false);
				rt.show(title);
			}
		}
		

		private double[] calculateXValues(double[] Vector, String plot){
			
			int N=Vector.length;
			double[]xValues = new double[N];
			
			if(plot=="MTF"){
				xValues[0]=0;
				//Scale of values for x-axis
				for(int i=1;i<N;i++){
					//xValues[i]=xValues[i-1]+(0.5/(N-1));
					xValues[i]=xValues[i-1]+(ny/(N-1));
				}
			}else{
				for(int i=0;i<N;i++){ 
					xValues[i]=i+1;
				}
			}
			return xValues;
		}
		
		////////////////Fourier transform module///////////////////
		
		//data type conversion from complex to double, to implement fft
		private double[] fftConversion(double[] Vector, String plot){
	        
			//Only half are necessary
			int N=Vector.length;
			int M=Vector.length/2;
			double divisor;
			Complex[] ArrayComplex = new Complex[N];
			Complex[] VectorFFTC = new Complex[N];
			double[]VectorFFTD=new double[M];
					
							
	        for (int i = 0; i < N; i++){ 
				//A double array is converted into a complex array ; imaginary part=0           
	            ArrayComplex[i] = new Complex(Vector[i], 0); 
	        }
			//FFT operation
	        VectorFFTC = fft(ArrayComplex); 
			
			
			if(plot=="SPP"){
				//Reject the first one
				for(int i = 1; i < M; i++){
					//absolute value (module)
					VectorFFTD[i-1]=VectorFFTC[i].abs()/VectorFFTC[1].abs();
				}
			}else{
				for (int i = 0; i < M; i++){
					//absolute value (module)
	            	VectorFFTD[i]=VectorFFTC[i].abs()/VectorFFTC[0].abs(); 
				}
	        }
			
			//Normalize
			if(plot=="SPP"){
				divisor=findMaxSPP(VectorFFTD);
			}else{			
				divisor=VectorFFTD[0];	
			}
			
			for (int i = 0; i < M; i++){
	           	VectorFFTD[i]=VectorFFTD[i]/divisor; 
	        }
			return VectorFFTD;
	    }
		
		
		private double findMaxSPP(double[] Vector){
			double max = 0;
	        for (int i=0; i<Vector.length; i++) {
	            if (Vector[i]>max) max=Vector[i];
	        }
			
			return max;
		}
		

		private class Complex {
	    	private final double re;   //the real part
	    	private final double im;   //the imaginary part

	    	// create a new object with the given real and imaginary parts
	    	public Complex(double real, double imag) {
	        	this.re = real;
	        	this.im = imag;
	    	}
			
			// return a string representation of the invoking object
		    public String toString()  {return re + " + " + im + "i"; }
			
			// return a new object whose value is (this + b)
	    	public Complex plus(Complex b) { 
	        	Complex a = this;             // invoking object
	        	double real = a.re + b.re;
	        	double imag = a.im + b.im;
	        	Complex sum = new Complex(real, imag);
	        	return sum;
	    	}
			
	    	public Complex minus(Complex b) { 
	        	Complex a = this;   
	        	double real = a.re - b.re;
	        	double imag = a.im - b.im;
	        	Complex diff = new Complex(real, imag);
	        	return diff;
	    	}		
			
			// return a new object whose value is (this * b)
	    	public Complex times(Complex b) {
	        	Complex a = this;
	        	double real = a.re * b.re - a.im * b.im;
	        	double imag = a.re * b.im + a.im * b.re;
	        	Complex prod = new Complex(real, imag);
	        	return prod;
	    	}
			
			// return |this|
	    	public double abs() { return Math.sqrt(re*re + im*im);  }
		}	
		
		
			public Complex[] fft(Complex[] x) {
	        
	        	int N = x.length;
	        	Complex[] y = new Complex[N];

		        // base case
	    	    if (N == 1) {
	        	    y[0] = x[0];
	            	return y;
	        	}

	        	// radix 2 Cooley-Tukey FFT
	        	if (N % 2 != 0) throw new RuntimeException("N is not a power of 2");
	        	Complex[] even = new Complex[N/2];
	        	Complex[] odd  = new Complex[N/2];
	        	for (int k = 0; k < N/2; k++) even[k] = x[2*k];
	        	for (int k = 0; k < N/2; k++) odd[k]  = x[2*k + 1];

		        Complex[] q = fft(even);
	        	Complex[] r = fft(odd);

	        	for (int k = 0; k < N/2; k++) {
	            	double kth = -2 * k * Math.PI / N;
	            	Complex wk = new Complex(Math.cos(kth), Math.sin(kth));
	            	y[k]       = q[k].plus(wk.times(r[k]));
	            	y[k + N/2] = q[k].minus(wk.times(r[k]));
	        	}
	        	return y;
	    	}
	}
	
}//class